@@ -75,7 +75,7 @@
 				     "cache/"
 				     (file-name-nondirectory f))))
 	     (directory-files (elmo-cache-folder-directory-internal folder)
-			      t "^[^.].*+"))))
+			      t "^[01][0-9A-F]$"))))
 
 (luna-define-method elmo-folder-message-file-p ((folder elmo-cache-folder))
   t)
@@ -87,48 +87,33 @@
    (elmo-cache-folder-directory-internal folder)))
 
 (luna-define-method elmo-folder-msgdb-create ((folder elmo-cache-folder)
-					      numbers new-mark
-					      already-mark seen-mark
-					      important-mark
-					      seen-list)
+					      numbers flag-table)
   (let ((i 0)
 	(len (length numbers))
-	overview number-alist mark-alist entity message-id
-	num mark)
+	(new-msgdb (elmo-make-msgdb))
+	entity message-id flags)
     (message "Creating msgdb...")
     (while numbers
       (setq entity
-	    (elmo-msgdb-create-overview-entity-from-file
+	    (elmo-msgdb-create-message-entity-from-file
+	     (elmo-msgdb-message-entity-handler new-msgdb)
 	     (car numbers) (elmo-message-file-name folder (car numbers))))
-      (if (null entity)
-	  ()
-	(setq num (elmo-msgdb-overview-entity-get-number entity))
-	(setq overview
-	      (elmo-msgdb-append-element
-	       overview entity))
-	(setq message-id (elmo-msgdb-overview-entity-get-id entity))
-	(setq number-alist
-	      (elmo-msgdb-number-add number-alist
-				     num
-				     message-id))
-	(if (setq mark (or (elmo-msgdb-global-mark-get message-id)
-			   (if (member message-id seen-list) nil new-mark)))
-	    (setq mark-alist
-		  (elmo-msgdb-mark-append
-		   mark-alist
-		   num mark)))
-	(when (> len elmo-display-progress-threshold)
-	  (setq i (1+ i))
-	  (elmo-display-progress
-	   'elmo-cache-folder-msgdb-create "Creating msgdb..."
-	   (/ (* i 100) len))))
+      (when entity
+	(setq message-id (elmo-message-entity-field entity 'message-id)
+	      flags (elmo-flag-table-get flag-table message-id))
+	(elmo-global-flags-set flags folder (car numbers) message-id)
+	(elmo-msgdb-append-entity new-msgdb entity flags))
+      (when (> len elmo-display-progress-threshold)
+	(setq i (1+ i))
+	(elmo-display-progress
+	 'elmo-cache-folder-msgdb-create "Creating msgdb..."
+	 (/ (* i 100) len)))
       (setq numbers (cdr numbers)))
     (message "Creating msgdb...done")
-    (list overview number-alist mark-alist)))
+    new-msgdb))
 
 (luna-define-method elmo-folder-append-buffer ((folder elmo-cache-folder)
-					       unread
-					       &optional number)
+					       &optional flags number)
   ;; dir-name is changed according to msgid.
   (unless (elmo-cache-folder-dir-name-internal folder)
     (let ((msgid (std11-field-body "message-id"))
@@ -142,6 +127,7 @@
 		   (not (file-exists-p file)))
 	  (write-region-as-binary
 	   (point-min) (point-max) file nil 'no-msg)
+	  (elmo-folder-preserve-flags folder msgid flags)
 	  t)))))
 
 (luna-define-method elmo-map-folder-delete-messages ((folder elmo-cache-folder)
@@ -149,7 +135,8 @@
   (dolist (location locations)
     (elmo-file-cache-delete
      (expand-file-name location
-		       (elmo-cache-folder-directory-internal folder)))))
+		       (elmo-cache-folder-directory-internal folder))))
+  t)
 
 (luna-define-method elmo-message-fetch-with-cache-process
   ((folder elmo-cache-folder) number strategy &optional section unseen)
@@ -174,27 +161,6 @@
 (luna-define-method elmo-message-file-p ((folder elmo-cache-folder) number)
   t)
 
-;;; To override elmo-map-folder methods.
-(luna-define-method elmo-folder-list-unreads-internal
-  ((folder elmo-cache-folder) unread-marks &optional mark-alist)
-  t)
-
-(luna-define-method elmo-folder-unmark-important ((folder elmo-cache-folder)
-						  numbers)
-  t)
-
-(luna-define-method elmo-folder-mark-as-important ((folder elmo-cache-folder)
-						   numbers)
-  t)
-
-(luna-define-method elmo-folder-unmark-read ((folder elmo-cache-folder)
-					     numbers)
-  t)
-
-(luna-define-method elmo-folder-mark-as-read ((folder elmo-cache-folder)
-					      numbers)
-  t)
-
 (require 'product)
 (product-provide (provide 'elmo-cache) (require 'elmo-version))
 
