@@ -63,20 +63,26 @@ Automatically loaded/saved.")
     (setq elmo-dop-queue (nconc elmo-dop-queue (list queue)))))
 
 (defvar elmo-dop-queue-merge-method-list
-  '(elmo-folder-mark-as-read
-    elmo-folder-unmark-read
-    elmo-folder-mark-as-important
-    elmo-folder-unmark-important))
+  '(elmo-folder-set-read-delayed
+    elmo-folder-unset-read-delayed
+    elmo-folder-set-important-delayed
+    elmo-folder-unset-important-delayed
+    elmo-folder-set-answered-delayed
+    elmo-folder-unset-answered-delayed))
 
 (defvar elmo-dop-queue-method-name-alist
   '((elmo-folder-append-buffer-dop-delayed . "Append")
     (elmo-folder-delete-messages-dop-delayed . "Delete")
     (elmo-message-encache . "Encache")
     (elmo-folder-create-dop-delayed . "Create")
-    (elmo-folder-mark-as-read . "Read")
-    (elmo-folder-unmark-read . "Unread")
-    (elmo-folder-mark-as-important . "Important")
-    (elmo-folder-unmark-important . "Unimportant")))
+    (elmo-folder-set-read-delayed . "Read")
+    (elmo-folder-unset-read-delayed . "Unread")
+    (elmo-folder-set-answered-delayed . "Answered")
+    (elmo-folder-unset-answered-delayed . "Unanswered")
+    (elmo-folder-set-important-delayed . "Important")
+    (elmo-folder-unset-important-delayed . "Unimportant")
+    (elmo-folder-set-flag . "Set flag")
+    (elmo-folder-unset-flag . "Unset flag")))
 
 (defmacro elmo-dop-queue-method-name (queue)
   `(cdr (assq (elmo-dop-queue-method ,queue)
@@ -191,21 +197,16 @@ which is corresponded to the FOLDER."
       (concat "+" (expand-file-name "spool" (elmo-folder-msgdb-path
 					     (, folder)))))))
 
-(defun elmo-dop-spool-folder-append-buffer (folder)
+(defun elmo-dop-spool-folder-append-buffer (folder flags)
   "Append current buffer content to the dop spool folder.
 FOLDER is the folder structure.
 Return a message number."
-  (setq folder (elmo-dop-spool-folder folder))
-  (unless (elmo-folder-exists-p folder)
-    (elmo-folder-create folder))
-  (let ((new-number (1+ (car (elmo-folder-status folder)))))
-    ;; dop folder is a localdir folder.
-    (write-region-as-binary (point-min) (point-max)
-			  (expand-file-name
-			   (int-to-string new-number)
-			   (elmo-localdir-folder-directory-internal folder))
-			  nil 'no-msg)
-    new-number))
+  (let ((spool (elmo-dop-spool-folder folder)))
+    (unless (elmo-folder-exists-p spool)
+      (elmo-folder-create spool))
+    (let ((new-number (1+ (car (elmo-folder-status spool)))))
+      (elmo-folder-append-buffer spool flags new-number)
+      new-number)))
 
 
 (defun elmo-dop-spool-folder-list-messages (folder)
@@ -229,13 +230,15 @@ FOLDER is the folder structure."
 				 'car
 				 (car (elmo-dop-queue-arguments queue)))))))))
 
+(defun elmo-dop-filter-pending-messages (numbers)
+  (elmo-delete-if (lambda (number) (< number 0)) numbers))
+
 ;;; DOP operations.
-(defsubst elmo-folder-append-buffer-dop (folder unread &optional number)
+(defsubst elmo-folder-append-buffer-dop (folder &optional flag number)
   (elmo-dop-queue-append
    folder 'elmo-folder-append-buffer-dop-delayed
-   (list unread
-	 (elmo-dop-spool-folder-append-buffer
-	  folder)
+   (list flag
+	 (elmo-dop-spool-folder-append-buffer folder flag)
 	 number)))
 
 (defsubst elmo-folder-delete-messages-dop (folder numbers)
@@ -262,96 +265,145 @@ FOLDER is the folder structure."
 (defsubst elmo-folder-create-dop (folder)
   (elmo-dop-queue-append folder 'elmo-folder-create-dop-delayed nil))
 
-(defsubst elmo-folder-mark-as-read-dop (folder numbers)
-  (elmo-dop-queue-append folder 'elmo-folder-mark-as-read (list numbers)))
+(defsubst elmo-folder-set-flag-dop (folder numbers flag)
+  (when (setq numbers (elmo-dop-filter-pending-messages numbers))
+    (let ((method (case flag
+		    (unread
+		     'elmo-folder-unset-read-delayed)
+		    (read
+		     'elmo-folder-set-read-delayed)
+		    (important
+		     'elmo-folder-set-important-delayed)
+		    (answered
+		     'elmo-folder-set-answered-delayed))))
+      (if method
+	  (elmo-dop-queue-append folder method (list numbers))
+	(elmo-dop-queue-append folder 'elmo-folder-set-flag
+			       (list numbers flag))))))
 
-(defsubst elmo-folder-unmark-read-dop (folder numbers)
-  (elmo-dop-queue-append folder 'elmo-folder-unmark-read (list numbers)))
-
-(defsubst elmo-folder-mark-as-important-dop (folder numbers)
-  (elmo-dop-queue-append folder 'elmo-folder-mark-as-important (list numbers)))
-
-(defsubst elmo-folder-unmark-important-dop (folder numbers)
-  (elmo-dop-queue-append folder 'elmo-folder-unmark-important (list numbers)))
+(defsubst elmo-folder-unset-flag-dop (folder numbers flag)
+  (when (setq numbers (elmo-dop-filter-pending-messages numbers))
+    (let ((method (case flag
+		    (unread
+		     'elmo-folder-set-read-delayed)
+		    (read
+		     'elmo-folder-unset-read-delayed)
+		    (important
+		     'elmo-folder-unset-important-delayed)
+		    (answered
+		     'elmo-folder-unset-answered-delayed))))
+      (if method
+	  (elmo-dop-queue-append folder method (list numbers))
+	(elmo-dop-queue-append folder 'elmo-folder-unset-flag
+			       (list numbers flag))))))
 
 ;;; Execute as subsutitute for plugged operation.
 (defun elmo-folder-status-dop (folder)
-  (let* ((number-alist (elmo-msgdb-number-load
-			(elmo-folder-msgdb-path folder)))
-	 (number-list (mapcar 'car number-alist))
-	 (spool-folder (elmo-dop-spool-folder folder))
-	 spool-length
-	 (i 0)
-	 max-num)
-    (setq spool-length
-	  (or (car (if (elmo-folder-exists-p spool-folder)
-		       (elmo-folder-status spool-folder)))
-	      0))
-    (setq max-num
-	  (or (nth (max (- (length number-list) 1) 0) number-list)
-	      0))
+  (let ((number-list (elmo-folder-list-messages folder nil 'in-msgdb))
+	(spool-folder (elmo-dop-spool-folder folder))
+	spool-length
+	max-num)
+    (setq spool-length (or (if (elmo-folder-exists-p spool-folder)
+			       (car (elmo-folder-status spool-folder)))
+			   0))
+    (setq max-num (if number-list (apply #'max number-list) 0))
     (cons (+ max-num spool-length) (+ (length number-list) spool-length))))
 
+(defun elmo-folder-next-message-number-dop (folder)
+  (let ((number-list (elmo-folder-list-messages folder nil 'in-msgdb))
+	(spool-folder (elmo-dop-spool-folder folder))
+	spool-length
+	max-num)
+    (setq spool-length (or (if (elmo-folder-exists-p spool-folder)
+			       (car (elmo-folder-status spool-folder)))
+			   0))
+    (setq max-num (if number-list (apply #'max number-list) 0))
+    (+ max-num spool-length)))
+
 ;;; Delayed operation (executed at online status).
-(defun elmo-folder-append-buffer-dop-delayed (folder unread number set-number)
+(defun elmo-folder-append-buffer-dop-delayed (folder flag number set-number)
   (let ((spool-folder (elmo-dop-spool-folder folder))
-	failure saved dequeued)
+	flags)
     (with-temp-buffer
-      (if (elmo-message-fetch spool-folder number
-			      (elmo-make-fetch-strategy 'entire)
-			      nil (current-buffer) 'unread)
-	  (condition-case nil
-	      (setq failure (not
-			     (elmo-folder-append-buffer
-			      folder unread set-number)))
-	    (error (setq failure t)))
-	(setq dequeued t)) ; Already deletef from queue.
-      (when failure
-	;; Append failed...
-	(setq saved (elmo-folder-append-buffer
-		     (elmo-make-folder elmo-lost+found-folder)
-		     unread set-number)))
-      (if (and (not dequeued)    ; if dequeued, no need to delete.
-	       (or (not failure) ; succeed
-		   saved))       ; in lost+found
-	  (elmo-folder-delete-messages spool-folder (list number)))
-      t)))
+      (when (elmo-message-fetch spool-folder number
+				(elmo-make-fetch-strategy 'entire)
+				nil (current-buffer) 'unread)
+	(setq flags (or (elmo-message-flags-for-append folder (* -1 number))
+			(cond ((listp flag)
+			       flag)
+			      ;; for compatibility with 2.11.12 or earlier
+			      ((eq flag t)
+			       nil)
+			      (t
+			       (list flag)))))
+	(when (or (condition-case nil
+		      (elmo-folder-append-buffer folder flags set-number)
+		    (error))
+		  ;; Append failed...
+		  (elmo-folder-append-buffer
+		   (elmo-make-folder elmo-lost+found-folder)
+		   flags))
+	  (elmo-folder-delete-messages spool-folder (list number)))))
+    ;; ignore failure (already dequed)
+    t))
 
 (defun elmo-folder-delete-messages-dop-delayed (folder number-alist)
-  (elmo-folder-delete-messages
-   folder
-   ;; messages are deleted only if message-id is not changed.
-   (mapcar 'car
-	   (elmo-delete-if
-	    (lambda (pair)
-	      (not (string=
-		    (cdr pair)
-		    (elmo-message-fetch-field folder (car pair)
-					      'message-id))))
-	    number-alist))))
+  (ignore-errors
+    (elmo-folder-delete-messages
+     folder
+     ;; messages are deleted only if message-id is not changed.
+     (mapcar 'car
+	     (elmo-delete-if
+	      (lambda (pair)
+		(not (string=
+		      (cdr pair)
+		      (elmo-message-fetch-field folder (car pair)
+						'message-id))))
+	      number-alist)))
+    t)) ; Always success (If failure, just remain)
 
 (defun elmo-folder-create-dop-delayed (folder)
   (unless (elmo-folder-exists-p folder)
     (elmo-folder-create folder)))
 
+(defun elmo-folder-set-important-delayed (folder numbers)
+  (elmo-folder-set-flag folder numbers 'important))
+
+(defun elmo-folder-unset-important-delayed (folder numbers)
+  (elmo-folder-unset-flag folder numbers 'important))
+
+(defun elmo-folder-set-read-delayed (folder numbers)
+  (elmo-folder-unset-flag folder numbers 'unread))
+
+(defun elmo-folder-unset-read-delayed (folder numbers)
+  (elmo-folder-set-flag folder numbers 'unread))
+
+(defun elmo-folder-set-answered-delayed (folder numbers)
+  (elmo-folder-set-flag folder numbers 'answered))
+
+(defun elmo-folder-unset-answered-delayed (folder numbers)
+  (elmo-folder-unset-flag folder numbers 'answered))
+
 ;;; Util
 (defun elmo-dop-msgdb (msgdb)
-  (list (mapcar (function
-		 (lambda (x)
-		   (elmo-msgdb-overview-entity-set-number
-		    x
-		    (* -1
-		       (elmo-msgdb-overview-entity-get-number x)))))
-		(nth 0 msgdb))
-	(mapcar (function
-		 (lambda (x) (cons
-			      (* -1 (car x))
-			      (cdr x))))
-		(nth 1 msgdb))
-	(mapcar (function
-		 (lambda (x) (cons
-			      (* -1 (car x))
-			      (cdr x)))) (nth 2 msgdb))))
+  (let ((new-db (elmo-make-msgdb))
+	entity)
+    (dolist (dop-entity (mapcar
+			 (lambda (number)
+			   (setq entity (elmo-msgdb-message-entity
+					 msgdb number))
+			   (elmo-message-entity-set-number
+			    entity
+			    (* -1 (elmo-message-entity-number entity)))
+			   entity)
+			 (elmo-msgdb-list-messages msgdb)))
+      (elmo-msgdb-append-entity new-db
+				dop-entity
+				(elmo-msgdb-flags
+				 msgdb
+				 (abs (elmo-message-entity-number
+				       dop-entity)))))
+    new-db))
 
 (require 'product)
 (product-provide (provide 'elmo-dop) (require 'elmo-version))
