@@ -44,6 +44,7 @@
 (defvar elmo-message-ignored-field-list mime-view-ignored-field-list)
 (defvar elmo-message-visible-field-list mime-view-visible-field-list)
 (defvar elmo-message-sorted-field-list nil)
+(defvar elmo-mime-display-header-analysis t)
 
 (defcustom elmo-mime-header-max-column fill-column
   "*Header max column number. Default is `fill-colmn'.
@@ -53,12 +54,6 @@ value is used."
 		 (function :tag "Function"))
   :group 'elmo)
 
-(defcustom elmo-mime-display-as-is-coding-system (if (boundp 'MULE)
-						     '*autoconv* 'undecided)
-  "*Coding system used when message is displayed as is."
-  :type 'symbol
-  :group 'elmo)
-
 (luna-define-method initialize-instance :after ((entity mime-elmo-buffer-entity)
 						&rest init-args)
   entity)
@@ -93,8 +88,11 @@ value is used."
 	    (setq field (intern
 			 (capitalize (buffer-substring f-b (1- p))))
 		  field-body (buffer-substring p f-e)
-		  field-decoder (inline (mime-find-field-decoder-internal
-					 field mode-obj)))
+		  field-decoder
+		  (if elmo-mime-display-header-analysis
+		      (inline (mime-find-field-decoder-internal
+			       field mode-obj))
+		    (inline (lambda (x y z) x))))
 	    (setq vf-alist (append (list
 				    (cons field-name
 					  (list field-body field-decoder)))
@@ -213,7 +211,7 @@ value is used."
 
 (defun elmo-mime-message-display (folder number viewbuf rawbuf original-mode
 					 &optional ignore-cache unread keymap)
-  "Display MIME message. 
+  "Display MIME message.
 A message in the FOLDER with NUMBER is displayed on the VIEWBUF using RAWBUF.
 VIEWBUF is a view buffer and RAWBUF is a raw buffer.
 ORIGINAL is the major mode of RAWBUF.
@@ -224,11 +222,11 @@ Return non-nil if not entire message was fetched."
   (let (mime-display-header-hook ; Do nothing.
 	(elmo-message-displaying t)
 	entity strategy)
-    (setq entity (elmo-msgdb-overview-get-entity number
-						 (elmo-folder-msgdb
-						  folder)))
-    (setq strategy (elmo-find-fetch-strategy folder entity
-					     ignore-cache))
+    (unless (zerop (elmo-folder-length folder))
+      (setq entity (elmo-message-entity folder number)))
+    (setq strategy (if entity (elmo-find-fetch-strategy folder entity
+							ignore-cache)
+		     (elmo-make-fetch-strategy 'entire)))
     (mime-display-message
      (mime-open-entity
       (if (and strategy
@@ -246,7 +244,7 @@ Return non-nil if not entire message was fetched."
 
 (defun elmo-mime-display-as-is (folder number viewbuf rawbuf original-mode
 				       &optional ignore-cache unread keymap)
-  "Display MIME message. 
+  "Display MIME message.
 A message in the FOLDER with NUMBER is displayed on the VIEWBUF using RAWBUF.
 VIEWBUF is a view buffer and RAWBUF is a raw buffer.
 ORIGINAL is the major mode of RAWBUF.
@@ -254,24 +252,35 @@ If optional argument IGNORE-CACHE is specified, existing cache is ignored.
 If second optional argument UNREAD is specified, message is displayed but
 keep it as unread.
 Return non-nil if cache is used."
-  (let ((entity (elmo-msgdb-overview-get-entity number
-						(elmo-folder-msgdb folder)))
-	mime-display-header-hook ; Do nothing.
-	cache-file strategy use-cache)
-    (setq cache-file (elmo-file-cache-get
-		      (elmo-msgdb-overview-entity-get-id entity)))
-    (setq use-cache (eq (elmo-file-cache-status cache-file) 'entire))
-    (setq strategy (elmo-make-fetch-strategy
-		    'entire use-cache (elmo-message-use-cache-p folder number)
-		    (elmo-file-cache-path
-		     cache-file)))
+  (let (mime-display-header-hook ; Do nothing.
+	(elmo-message-displaying t)
+	entity cache-file cache-used)
+    (unless (zerop (elmo-folder-length folder))
+      (setq entity (elmo-message-entity folder number)))
+    (when entity
+      (setq cache-file
+	    (elmo-file-cache-get
+	     (elmo-message-entity-field entity 'message-id)))
+      ;; Required to be an entire cache.
+      (unless (eq (elmo-file-cache-status cache-file) 'entire)
+	(setq ignore-cache t)))
     (elmo-mime-display-as-is-internal
      (mime-open-entity
       'elmo-buffer
       (elmo-make-mime-message-location
-       folder number strategy rawbuf unread))
+       folder number
+       (elmo-make-fetch-strategy 'entire
+				 (unless ignore-cache
+				   (setq
+				    cache-used
+				    (and cache-file
+					 (elmo-file-cache-status cache-file))))
+				 (elmo-message-use-cache-p folder number)
+				 (and cache-file
+				      (elmo-file-cache-path cache-file)))
+       rawbuf unread))
      viewbuf nil keymap original-mode)
-    (elmo-fetch-strategy-use-cache strategy)))
+    cache-used))
 
 ;; Replacement of mime-display-message.
 (defun elmo-mime-display-as-is-internal (message
@@ -297,28 +306,32 @@ Return non-nil if cache is used."
 
       ;; Humm...
       (set-buffer-multibyte nil)
-      (mime-insert-entity message)
+      (insert (mime-entity-body message))
       (set-buffer-multibyte t)
       (decode-coding-region (point-min) (point-max)
 			    elmo-mime-display-as-is-coding-system)
-      (save-restriction
-	(std11-narrow-to-header)
-	(run-hooks 'elmo-message-header-inserted-hook))
+      (goto-char (point-min))
+      (insert "\n")
+      (goto-char (point-min))
+
+      (let ((method (cdr (assq original-major-mode
+			       mime-header-presentation-method-alist))))
+	(if (functionp method)
+	    (funcall method message nil)))
+
       ;; set original major mode for mime-preview-quit
       (put-text-property (point-min) (point-max)
 			 'mime-view-situation
 			 `((major-mode . ,original-major-mode)))
+      (put-text-property (point-min) (point-max)
+			 'elmo-as-is-entity message)
       (use-local-map
        (or keymap
 	   (if default-keymap-or-function
 	       (mime-view-define-keymap default-keymap-or-function)
 	     mime-view-mode-default-map)))
-      (let ((point
-	     (next-single-property-change (point-min) 'mime-view-entity)))
-	(if point
-	    (goto-char point)
-	  (goto-char (point-min))
-	  (search-forward "\n\n" nil t)))
+      (goto-char (point-min))
+      (search-forward "\n\n" nil t)
       (run-hooks 'mime-view-mode-hook)
       (set-buffer-modified-p nil)
       (setq buffer-read-only t)
