@@ -120,10 +120,10 @@
      wl-summary-jump-to-current-message t "Jump to Current Message"]
     [wl-summary-sync-force-update
      wl-summary-sync-force-update t "Sync Current Folder"]
-    [wl-summary-delete
-     wl-summary-delete t "Delete Current Message"]
-    [wl-summary-mark-as-important
-     wl-summary-mark-as-important t "Mark Current Message as Important"]
+    [wl-summary-dispose
+     wl-summary-dispose t "Dispose Current Message"]
+    [wl-summary-set-flags
+     wl-summary-set-flags t "Set Flags"]
     [wl-draft
      wl-summary-write-current-folder t "Write for Current Folder"]
     [wl-summary-reply
@@ -171,7 +171,7 @@
 
 (eval-when-compile
   (defmacro wl-e21-display-image-p ()
-    '(and (display-images-p)
+    '(and wl-highlight-folder-with-icon
 	  (image-type-available-p 'xpm))))
 
 (defun wl-e21-setup-toolbar (bar)
@@ -248,33 +248,31 @@
 
 (eval-when-compile
   (defsubst wl-e21-highlight-folder-group-line (start end icon numbers)
-    (when (wl-e21-display-image-p)
-      (let (overlay)
-	(let ((overlays (overlays-in start end)))
-	  (while (and (setq overlay (pop overlays))
-		      (not (overlay-get overlay 'wl-e21-icon)))))
-	(unless overlay
-	  (setq overlay (make-overlay start end))
-	  (overlay-put overlay 'wl-e21-icon t)
-	  (overlay-put overlay 'evaporate t))
-	(let ((image (get icon 'image)))
+    (let (image)
+      (when (wl-e21-display-image-p)
+	(let (overlay)
+	  (let ((overlays (overlays-in start end)))
+	    (while (and (setq overlay (pop overlays))
+			(not (overlay-get overlay 'wl-e21-icon)))))
+	  (unless overlay
+	    (setq overlay (make-overlay start end))
+	    (overlay-put overlay 'wl-e21-icon t)
+	    (overlay-put overlay 'evaporate t))
+	  (setq image (get icon 'image))
 	  (unless image
 	    (let ((name (symbol-value
 			 (cdr (assq icon wl-folder-toggle-icon-list))))
 		  (load-path (cons wl-icon-directory load-path)))
-	      (when (setq image (find-image `((:type xpm :file ,name
-						     :ascent center))))
-		(setq image (put icon 'image (propertize name
-							 'display image))))))
-	  (overlay-put overlay 'before-string image)
-	  (overlay-put overlay 'invisible (and image t))
-	  (when (and wl-use-highlight-mouse-line (display-mouse-p))
-	    (let ((inhibit-read-only t))
-	      (put-text-property (if image
-				     (max (1- start) (line-beginning-position))
-				   start)
-				 (line-end-position)
-				 'mouse-face 'highlight)))))))
+	      (setq image (find-image `((:type xpm :file ,name
+					       :ascent center))))))
+	  (overlay-put overlay 'display image)))
+      (when (and wl-use-highlight-mouse-line (display-mouse-p))
+	(let ((inhibit-read-only t))
+	  (put-text-property (if image
+				 (max (1- start) (line-beginning-position))
+			       start)
+			     (line-end-position)
+			     'mouse-face 'highlight)))))
 
   (defsubst wl-e21-highlight-folder-by-numbers (start end text-face numbers)
     (when (display-color-p)
@@ -311,7 +309,8 @@
     (let (fld-name start end)
       (cond
        (;; opened folder group
-	(looking-at wl-highlight-folder-opened-regexp)
+	(and (wl-folder-buffer-group-p)
+	     (looking-at wl-highlight-folder-opened-regexp))
 	(setq start (match-beginning 1)
 	      end (match-end 1))
 	(wl-e21-highlight-folder-group-line start end
@@ -321,7 +320,8 @@
 					    'wl-highlight-folder-opened-face
 					    numbers))
        (;; closed folder group
-	(looking-at wl-highlight-folder-closed-regexp)
+	(and (wl-folder-buffer-group-p)
+	     (looking-at wl-highlight-folder-closed-regexp))
 	(setq start (match-beginning 1)
 	      end (match-end 1))
 	(wl-e21-highlight-folder-group-line start end
@@ -361,10 +361,14 @@
 			    ((string= fld-name wl-queue-folder);; queue folder
 			     (get 'wl-folder-queue-image 'image))
 			    (;; and one of many other folders
-			     (setq type (elmo-folder-type fld-name))
+			     (setq type (or (elmo-folder-type fld-name)
+					    (elmo-folder-type-internal
+					     (elmo-make-folder fld-name))))
 			     (get (intern (format "wl-folder-%s-image" type))
 				  'image)))))
-	      (overlay-put overlay 'before-string image)))
+	      (overlay-put overlay 'before-string
+			   (propertize " " 'display image
+				       'invisible t))))
 	  (when (and wl-use-highlight-mouse-line (display-mouse-p))
 	    (let ((inhibit-read-only t))
 	      (put-text-property (if image
@@ -404,44 +408,43 @@
 		(setq overlay (make-overlay start end))
 		(overlay-put overlay 'wl-e21-icon t)
 		(overlay-put overlay 'evaporate t))
-	      (put-text-property 0 (length status) 'display image status)
-	      (overlay-put overlay 'before-string status)
-	      (overlay-put overlay 'invisible t))))))))
+	      (overlay-put overlay 'display image))))))))
 
 (defun wl-plugged-set-folder-icon (folder string)
-  (if (wl-e21-display-image-p)
-      (let (type)
-	(cond ((string= folder wl-queue-folder)
-	       (concat (get 'wl-folder-queue-image 'image)
-		       string))
-	      ((setq type (elmo-folder-type folder))
-	       (concat (get (intern (format "wl-folder-%s-image"
-					    type))
-			    'image)
-		       string))
-	      (t
-	       string)))
-    string))
+  (let (image type)
+    (when (wl-e21-display-image-p)
+      (setq image
+	    (cond ((string= folder wl-queue-folder)
+		   (get 'wl-folder-queue-image 'image))
+		  ((setq type (or (elmo-folder-type folder)
+				  (elmo-folder-type-internal
+				   (elmo-make-folder folder))))
+		   (get (intern (format "wl-folder-%s-image" type))
+			'image)))))
+    (if image
+	(concat (propertize " " 'display image 'invisible t) string)
+      string)))
 
 (defvar wl-folder-internal-icon-list
   ;; alist of (image . icon-file)
-  '((wl-folder-nntp-image         . wl-nntp-folder-icon)
-    (wl-folder-imap4-image        . wl-imap-folder-icon)
-    (wl-folder-pop3-image         . wl-pop-folder-icon)
+  '((wl-folder-nntp-image	  . wl-nntp-folder-icon)
+    (wl-folder-imap4-image	  . wl-imap-folder-icon)
+    (wl-folder-pop3-image	  . wl-pop-folder-icon)
     (wl-folder-localdir-image     . wl-localdir-folder-icon)
     (wl-folder-localnews-image    . wl-localnews-folder-icon)
     (wl-folder-internal-image     . wl-internal-folder-icon)
-    (wl-folder-multi-image        . wl-multi-folder-icon)
+    (wl-folder-multi-image	  . wl-multi-folder-icon)
     (wl-folder-filter-image       . wl-filter-folder-icon)
     (wl-folder-archive-image      . wl-archive-folder-icon)
-    (wl-folder-pipe-image         . wl-pipe-folder-icon)
+    (wl-folder-pipe-image	  . wl-pipe-folder-icon)
     (wl-folder-maildir-image      . wl-maildir-folder-icon)
-    (wl-folder-nmz-image          . wl-nmz-folder-icon)
+    (wl-folder-nmz-image	  . wl-nmz-folder-icon)
     (wl-folder-shimbun-image      . wl-shimbun-folder-icon)
+    (wl-folder-file-image	  . wl-file-folder-icon)
     (wl-folder-trash-empty-image  . wl-empty-trash-folder-icon)
-    (wl-folder-draft-image        . wl-draft-folder-icon)
-    (wl-folder-queue-image        . wl-queue-folder-icon)
-    (wl-folder-trash-image        . wl-trash-folder-icon)))
+    (wl-folder-draft-image	  . wl-draft-folder-icon)
+    (wl-folder-queue-image	  . wl-queue-folder-icon)
+    (wl-folder-trash-image	  . wl-trash-folder-icon)))
 
 (defun wl-folder-init-icons ()
   (when (wl-e21-display-image-p)
@@ -453,7 +456,7 @@
 	  (setq name (symbol-value (cdr icon))
 		image (find-image `((:type xpm :file ,name :ascent center))))
 	  (when image
-	    (put (car icon) 'image (propertize name 'display image))))))))
+	    (put (car icon) 'image image)))))))
 
 (defun wl-plugged-init-icons ()
   (let ((props (when (display-mouse-p)
@@ -529,10 +532,16 @@
   (let ((keymap (make-sparse-keymap)))
     (define-key keymap "D" 'wl-message-delete-current-part)
     (define-key keymap "l" 'wl-message-toggle-disp-summary)
+    (define-key keymap "\C-c:d" 'wl-message-decrypt-pgp-nonmime)
+    (define-key keymap "\C-c:v" 'wl-message-verify-pgp-nonmime)
+    (define-key keymap "w" 'wl-draft)
     (define-key keymap [mouse-4] 'wl-message-wheel-down)
     (define-key keymap [mouse-5] 'wl-message-wheel-up)
     (define-key keymap [S-mouse-4] 'wl-message-wheel-down)
     (define-key keymap [S-mouse-5] 'wl-message-wheel-up)
+    ;; Meadow2
+    (define-key keymap [mouse-wheel1] 'wl-message-wheel-dispatcher)
+    (define-key keymap [S-mouse-wheel1] 'wl-message-wheel-dispatcher)
     (set-keymap-parent wl-message-button-map keymap)
     (define-key wl-message-button-map
       [mouse-2] 'wl-message-button-dispatcher)
@@ -540,6 +549,23 @@
 
 (defalias 'wl-setup-message 'wl-e21-setup-message-toolbar)
 
+;; Wheel handling for Meadow2
+(defun wl-message-wheel-dispatcher (event)
+  (interactive "e")
+  (if (< (nth 4 (nth 1 event)) 0)
+      (wl-message-wheel-up event)
+    (wl-message-wheel-down event)))
+
+(defun wl-summary-wheel-dispatcher (event)
+  (interactive "e")
+  (if (< (nth 4 (nth 1 event)) 0)
+      (if (memq 'shift (event-modifiers event))
+	  (wl-summary-down)
+	(wl-summary-next))
+    (if (memq 'shift (event-modifiers event))
+	(wl-summary-up)
+      (wl-summary-prev))))
+
 (defun wl-message-wheel-up (event)
   (interactive "e")
   (if (string-match (regexp-quote wl-message-buffer-name)
@@ -626,7 +652,9 @@ Special commands:
 ;;  (define-key wl-draft-mode-map "\C-x\C-s" 'wl-draft-save)
   (define-key wl-draft-mode-map "\C-xk"    'wl-draft-mimic-kill-buffer)
   (define-key wl-draft-mode-map "\C-c\C-d" 'wl-draft-elide-region)
-  (define-key wl-draft-mode-map "\C-a" 'wl-draft-beginning-of-line))
+  (define-key wl-draft-mode-map "\C-a" 'wl-draft-beginning-of-line)
+  (define-key wl-draft-mode-map "\M-p" 'wl-draft-previous-history-element)
+  (define-key wl-draft-mode-map "\M-n" 'wl-draft-next-history-element))
 
 (defun wl-draft-overload-functions ()
   (wl-mode-line-buffer-identification)
@@ -641,6 +669,8 @@ Special commands:
   (let ((event (read-event)))
     (cons (and (numberp event) event) event)))
 
+(defalias 'wl-completing-read-multiple 'completing-read-multiple)
+
 (put 'wl-modeline-biff-state-on 'risky-local-variable t)
 (put 'wl-modeline-biff-state-off 'risky-local-variable t)
 (put 'wl-modeline-plug-state-on 'risky-local-variable t)
