@@ -164,6 +164,11 @@ If HACK-ADDRESSES is t, then the strings are considered to be mail addresses,
 	  (funcall func))
       (wl-push (cdr keve) unread-command-events))))
 
+(defun wl-require-update-all-folder-p (name)
+  "Return non-nil if NAME is draft or queue folder."
+  (or (string= name wl-draft-folder)
+      (string= name wl-queue-folder)))
+
 ;(defalias 'wl-make-hash 'elmo-make-hash)
 ;;(make-obsolete 'wl-make-hash 'elmo-make-hash)
 
@@ -202,7 +207,7 @@ even when invalid character is contained."
 	  (concat string paddings)))))
    (t
     (elmo-set-work-buf
-     (elmo-set-buffer-multibyte default-enable-multibyte-characters)
+     (set-buffer-multibyte default-enable-multibyte-characters)
      (insert string)
      (when (> (current-column) (abs width))
        (when (> (move-to-column (abs width)) (abs width))
@@ -260,13 +265,17 @@ even when invalid character is contained."
 	  value pair)
       (while alist
 	(setq pair (car alist))
-	(if (string-match (car pair) folder)
-	    (cond ((eq match 'all)
-		   (setq value (append value (list (cdr pair)))))
-		  ((eq match 'all-list)
-		   (setq value (append value (cdr pair))))
-		  ((not match)
-		   (throw 'found (cdr pair)))))
+	(if (and (eq match 'function)
+		 (functionp (car pair)))
+	    (when (funcall (car pair) folder)
+	      (throw 'found (cdr pair)))
+	  (if (string-match (car pair) folder)
+	      (cond ((eq match 'all)
+		     (setq value (append value (list (cdr pair)))))
+		    ((eq match 'all-list)
+		     (setq value (append value (cdr pair))))
+		    ((or (not match) (eq match 'function))
+		     (throw 'found (cdr pair))))))
 	(setq alist (cdr alist)))
       value)))
 
@@ -305,14 +314,6 @@ even when invalid character is contained."
 (defalias 'wl-string 'elmo-string)
 (make-obsolete 'wl-string 'elmo-string)
 
-;; Check if active region exists or not.
-(if (boundp 'mark-active)
-    (defmacro wl-region-exists-p ()
-      'mark-active)
-  (if (fboundp 'region-exists-p)
-      (defmacro wl-region-exists-p ()
-	(list 'region-exists-p))))
-
 (if (not (fboundp 'overlays-in))
     (defun overlays-in (beg end)
       "Return a list of the overlays that overlap the region BEG ... END.
@@ -345,14 +346,6 @@ or between BEG and END."
       (setq loop (- loop 1)))
     ret-val))
 
-(defun wl-list-diff (list1 list2)
-  "Return a list of elements of LIST1 that do not appear in LIST2."
-  (let ((list1 (copy-sequence list1)))
-    (while list2
-      (setq list1 (delq (car list2) list1))
-      (setq list2 (cdr list2)))
-    list1))
-
 (defun wl-append-assoc-list (item value alist)
   "make assoc list '((item1 value1-1 value1-2 ...)) (item2 value2-1 ...)))"
   (let ((entry (assoc item alist)))
@@ -531,19 +524,23 @@ that `read' can handle, whenever this is possible."
 	       folder nil nil nil t)
 	    (wl-summary-goto-folder-subr
 	     folder 'update nil nil t)
-	    (goto-char (point-min))
-	    (re-search-forward (concat "^ *" msg) nil t)
+	    (wl-summary-jump-to-msg (string-to-number msg))
 	    (wl-summary-redisplay)))
       (message "Not a nntp: url."))))
 
 (defmacro wl-concat-list (list separator)
   (` (mapconcat 'identity (delete "" (delq nil (, list))) (, separator))))
 
-(defmacro wl-current-message-buffer ()
-  (` (save-excursion
-       (if (buffer-live-p wl-current-summary-buffer)
-	   (set-buffer wl-current-summary-buffer))
-       wl-message-buffer)))
+(defun wl-current-message-buffer ()
+  (when (buffer-live-p wl-current-summary-buffer)
+    (with-current-buffer wl-current-summary-buffer
+      (or wl-message-buffer
+	  (and (wl-summary-message-number)
+	       (car (wl-message-buffer-display
+		     wl-summary-buffer-elmo-folder
+		     (wl-summary-message-number)
+		     wl-summary-buffer-display-mime-mode
+		     nil nil)))))))
 
 (defmacro wl-kill-buffers (regexp)
   (` (mapcar (function
@@ -569,20 +566,28 @@ that `read' can handle, whenever this is possible."
 
 (defun wl-collect-draft ()
   (let ((draft-regexp (concat
-		       "^" (regexp-quote
-			    (elmo-localdir-folder-directory-internal
-			     (wl-folder-get-elmo-folder wl-draft-folder)))))
+		       "^" (regexp-quote wl-draft-folder)))
 	result buf)
     (mapcar
      (function (lambda (x)
-		 (if (and
-		      (setq buf (with-current-buffer x
-				  wl-draft-buffer-file-name))
-		      (string-match draft-regexp buf))
+		 (if (with-current-buffer x
+		       (and (eq major-mode 'wl-draft-mode)
+			    (buffer-name)
+			    (string-match draft-regexp (buffer-name))))
 		     (setq result (nconc result (list x))))))
      (buffer-list))
     result))
 
+(defun wl-save-drafts ()
+  (let ((msg (current-message))
+	(buffers (wl-collect-draft)))
+    (save-excursion
+      (while buffers
+	(set-buffer (car buffers))
+	(if (buffer-modified-p) (wl-draft-save))
+	(setq buffers (cdr buffers))))
+    (message msg)))
+
 (static-if (fboundp 'read-directory-name)
     (defun wl-read-directory-name (prompt dir)
       (read-directory-name prompt dir dir))
@@ -712,6 +717,34 @@ that `read' can handle, whenever this is possible."
 	  (set-window-hscroll (get-buffer-window (current-buffer) t) 0))
 	max))))
 
+;; Draft auto-save
+(static-cond
+ (wl-on-xemacs
+  (defvar wl-save-drafts-timer-name "wl-save-drafts")
+
+  (defun wl-set-save-drafts ()
+    (if (numberp wl-auto-save-drafts-interval)
+	(unless (get-itimer wl-save-drafts-timer-name)
+	  (start-itimer wl-save-drafts-timer-name 'wl-save-drafts
+			wl-auto-save-drafts-interval wl-auto-save-drafts-interval
+			t))
+      (when (get-itimer wl-save-drafts-timer-name)
+	(delete-itimer wl-save-drafts-timer-name)))))
+ (t
+  (defun wl-set-save-drafts ()
+    (if (numberp wl-auto-save-drafts-interval)
+	(progn
+	  (require 'timer)
+	  (if (get 'wl-save-drafts 'timer)
+	      (progn (timer-set-idle-time (get 'wl-save-drafts 'timer)
+					  wl-auto-save-drafts-interval t)
+		     (timer-activate-when-idle (get 'wl-save-drafts 'timer)))
+	    (put 'wl-save-drafts 'timer
+		 (run-with-idle-timer
+		  wl-auto-save-drafts-interval t 'wl-save-drafts))))
+      (when (get 'wl-save-drafts 'timer)
+	(cancel-timer (get 'wl-save-drafts 'timer)))))))
+
 ;; Biff
 (static-cond
  (wl-on-xemacs
@@ -724,13 +757,11 @@ that `read' can handle, whenever this is possible."
   (defun wl-biff-start ()
     (wl-biff-stop)
     (when wl-biff-check-folder-list
-      (wl-biff-check-folders)
       (start-itimer wl-biff-timer-name 'wl-biff-check-folders
-		    wl-biff-check-interval wl-biff-check-interval))))
-
- ((and (condition-case nil (require 'timer) (error nil));; FSFmacs 19+
-       (fboundp 'timer-activate))
+		    wl-biff-check-interval wl-biff-check-interval
+		    wl-biff-use-idle-timer))))
 
+ (t
   (defun wl-biff-stop ()
     (when (get 'wl-biff 'timer)
       (cancel-timer (get 'wl-biff 'timer))))
@@ -738,14 +769,27 @@ that `read' can handle, whenever this is possible."
   (defun wl-biff-start ()
     (require 'timer)
     (when wl-biff-check-folder-list
-      (wl-biff-check-folders)
-      (if (get 'wl-biff 'timer)
-	  (timer-activate (get 'wl-biff 'timer))
-	(put 'wl-biff 'timer (run-at-time
+      (if wl-biff-use-idle-timer
+	  (if (get 'wl-biff 'timer)
+	      (progn (timer-set-idle-time (get 'wl-biff 'timer)
+					  wl-biff-check-interval t)
+		     (timer-activate-when-idle (get 'wl-biff 'timer)))
+	    (put 'wl-biff 'timer
+		 (run-with-idle-timer
+		  wl-biff-check-interval t 'wl-biff-event-handler)))
+	(if (get 'wl-biff 'timer)
+	    (progn
+	      (timer-set-time (get 'wl-biff 'timer)
 			      (timer-next-integral-multiple-of-time
 			       (current-time) wl-biff-check-interval)
-			      wl-biff-check-interval
-			      'wl-biff-event-handler)))))
+			      wl-biff-check-interval)
+	      (timer-activate (get 'wl-biff 'timer)))
+	  (put 'wl-biff 'timer
+	       (run-at-time
+		(timer-next-integral-multiple-of-time
+		 (current-time) wl-biff-check-interval)
+		wl-biff-check-interval
+		'wl-biff-event-handler))))))
 
   (defun-maybe timer-next-integral-multiple-of-time (time secs)
     "Yield the next value after TIME that is an integral multiple of SECS.
@@ -803,10 +847,7 @@ This function is imported from Emacs 20.7."
 	    (timer-set-time timer (timer-next-integral-multiple-of-time
 				   current wl-biff-check-interval)
 			    wl-biff-check-interval)
-	    (timer-activate timer))))))
- (t
-  (fset 'wl-biff-stop 'ignore)
-  (fset 'wl-biff-start 'ignore)))
+	    (timer-activate timer)))))))
 
 (defsubst wl-biff-notify (new-mails notify-minibuf)
   (when (and (not wl-modeline-biff-status) (> new-mails 0))
@@ -841,7 +882,8 @@ This function is imported from Emacs 20.7."
 	    (while flist
 	      (setq folder (wl-folder-get-elmo-folder (car flist))
 		    flist (cdr flist))
-	      (when (elmo-folder-plugged-p folder)
+	      (when (and (elmo-folder-plugged-p folder)
+			 (elmo-folder-exists-p folder))
 		(setq new-mails
 		      (+ new-mails
 			 (nth 0 (wl-biff-check-folder folder))))))
@@ -871,23 +913,26 @@ This function is imported from Emacs 20.7."
   (wl-biff-notify (car diff) (nth 2 data)))
 
 (defun wl-biff-check-folder-async (folder notify-minibuf)
-  (when (elmo-folder-plugged-p folder)
-    (elmo-folder-set-biff-internal folder t)
-    (if (and (eq (elmo-folder-type-internal folder) 'imap4)
-	     (elmo-folder-use-flag-p folder))
-	;; Check asynchronously only when IMAP4 and use server diff.
-	(progn
-	  (setq elmo-folder-diff-async-callback
-		'wl-biff-check-folder-async-callback)
-	  (setq elmo-folder-diff-async-callback-data
-		(list (elmo-folder-name-internal folder)
-		      (get-buffer wl-folder-buffer-name)
-		      notify-minibuf))
-	  (elmo-folder-diff-async folder))
-      (unwind-protect
-	  (wl-biff-notify (car (wl-biff-check-folder folder))
-			  notify-minibuf)
-	(setq wl-biff-check-folders-running nil)))))
+  (if (and (elmo-folder-plugged-p folder)
+	   (wl-folder-entity-exists-p (elmo-folder-name-internal folder)))
+      (progn
+	(elmo-folder-set-biff-internal folder t)
+	(if (and (eq (elmo-folder-type-internal folder) 'imap4)
+		 (elmo-folder-use-flag-p folder))
+	    ;; Check asynchronously only when IMAP4 and use server diff.
+	    (progn
+	      (setq elmo-folder-diff-async-callback
+		    'wl-biff-check-folder-async-callback)
+	      (setq elmo-folder-diff-async-callback-data
+		    (list (elmo-folder-name-internal folder)
+			  (get-buffer wl-folder-buffer-name)
+			  notify-minibuf))
+	      (elmo-folder-diff-async folder))
+	  (unwind-protect
+	      (wl-biff-notify (car (wl-biff-check-folder folder))
+			      notify-minibuf)
+	    (setq wl-biff-check-folders-running nil))))
+    (setq wl-biff-check-folders-running nil)))
 
 (if (and (fboundp 'regexp-opt)
 	 (not (featurep 'xemacs)))
@@ -902,13 +947,19 @@ is enclosed by at least one regexp grouping construct."
 	      close-paren))))
 
 (defalias 'wl-expand-newtext 'elmo-expand-newtext)
+(defalias 'wl-regexp-opt 'elmo-regexp-opt)
 
 (defun wl-region-exists-p ()
   "Return non-nil if a region exists on current buffer."
   (static-if (featurep 'xemacs)
-      (and zmacs-regions zmacs-region-active-p)
+      (region-active-p)
     (and transient-mark-mode mark-active)))
 
+(defun wl-deactivate-region ()
+  "Deactivate region on current buffer"
+  (static-if (not (featurep 'xemacs))
+      (setq mark-active nil)))
+
 (defvar wl-line-string)
 (defun wl-line-parse-format (format spec-alist)
   "Make a formatter from FORMAT and SPEC-ALIST."
@@ -984,6 +1035,14 @@ is enclosed by at least one regexp grouping construct."
        (when (get-buffer "*Compile-Log-Show*")
 	 (bury-buffer "*Compile-Log-Show*")))))
 
+(defsubst wl-copy-local-variables (src dst local-variables)
+  "Copy value of LOCAL-VARIABLES from SRC buffer to DST buffer."
+  (with-current-buffer dst
+    (dolist (variable local-variables)
+      (set (make-local-variable variable)
+	   (with-current-buffer src
+	     (symbol-value variable))))))
+
 (require 'product)
 (product-provide (provide 'wl-util) (require 'wl-version))
 
